# This script estimates the values 'Dx', 'Dz', 'Ly' and 'Lz' as used in the 'xyzab-srt' kinematic.
# Measurements must be placed in '/xyzab_srt.txt' in csv format (Angle A[°],Angle B[°],Z-joint position)
#
# Note: the length of the tool used for measuring is included in the value of 'Lz'. So for proper 
# function of the kinematic model the user needs to subtract the tool length (eg from ball tip to 
# spindle nose) from the result of 'Lz'

# Author: David Mueller

import csv, sys
from numpy import radians, array, sin, cos, sqrt, matrix, zeros, floor, identity, diag
from numpy.linalg import pinv, norm

def import_csv_to_list_of_dicts(file): # import csv data as list of dicts
    data_keys = ['A', 'B', 'Qz']  
    # Prepare a dict with the data keys and empty values
    template_dict = {}
    data = [] # our data will be a list of dicts
    z_offset = None
    with open(file, newline='') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            new_dict = template_dict.copy() 
            new_dict.update(row) 
            # we are looking for a measurement with A and B = zero to extract the absolute position of z=0
            if float(new_dict['A']) == 0 and float(new_dict['B']) == 0:
                z_offset = float(new_dict['Qz']) 
            # we need to change the values from string to float and also the angle from degrees to radians
            new_dict['A'] = radians(float(new_dict['A']))
            new_dict['B'] = radians(float(new_dict['B']))
            new_dict['Qz'] = float(new_dict['Qz'])
            print(new_dict)
            data.append(new_dict.copy())
    print(' ')
    print('Z-Offset: ',z_offset)
    return(data,z_offset)

def least_square_estimator(lod, z_offset):
    use_weights = 0
    # Assemble Z-Vector
    Z = []
    for l in range(len(lod)):
        Z.append([lod[l]['Qz'] - z_offset])
    Z = array(Z)

    print(' ')
    print('Z:')
    print(Z)

    # Assemble A-Matrix
    rows =  len(Z)
    cols = 4
    A = matrix(zeros([rows, cols]))
    for l in range(len(lod)):
        A[l,0] =     sin(lod[l]['B'])
        A[l,1] = 1 - cos(lod[l]['B'])
        A[l,2] =   - sin(lod[l]['A']) * cos(lod[l]['B'])
        A[l,3] = 1 - cos(lod[l]['A']) * cos(lod[l]['B'])

    print(' ')
    print('A:')
    print(A)

    # Calculate least square estimator 
    X = pinv(A.T*A)*A.T*Z
    print(' ')    
    print('Estimator X:')
    print(X)

    # Calculate error 
    V = A*X - Z
    print(' ')    
    print('V:')
    print(V)

    # Estimation of error variance 
    # ie the squared deviation from the mean
    sigma2 = (V.T*V)/(rows-cols)

    print(' ')    
    print('Error variance, \u03C3\u00b2:')
    print(sigma2)
    
    # Calculate variance-covariance matrix 
    # where the diagonal corresponds to the squared variances
    # of the estimated values
    SigmaX = sigma2[0,0]*pinv(A.T*A)
    print(' ')    
    print('Variance of estimated values, \u03A3X:')
    print(sqrt(diag(SigmaX)))

    Dx = round(float(X[0, 0]),4)
    Dz = round(float(X[1, 0]),4)
    Ly = round(float(X[2, 0]),4)
    Lz = round(float(X[3, 0]),4)
    s_Dx = round(float(sqrt(SigmaX[0, 0])),3)
    s_Dz = round(float(sqrt(SigmaX[1, 1])),3)
    s_Ly = round(float(sqrt(SigmaX[2, 2])),3)
    s_Lz = round(float(sqrt(SigmaX[3, 3])),3)

    return(Dx,Dz,Ly,Lz,s_Dx,s_Dz,s_Ly,s_Lz)




def main() :
    # Import the dataset
    file = './xyzab_srt.txt'
    (data, z_offset) = import_csv_to_list_of_dicts(file)
    # if no measurement at A=B=0 has been found in the data then we exit with an error
    if z_offset == None:
        print('\nERROR: No measurement at A,B = (0,0) found. Unable to calculate Z-Offset. ')
        sys.exit()
    # Run the least square estimation  
    (Dx,Dz,Ly,Lz,s_Dx,s_Dz,s_Ly,s_Lz) = least_square_estimator(data, z_offset)

    print(' ')
    print('Dx = ', -Dx, ' (\u03A3X:', s_Dx,')')
    print('Dz = ', -Dz, ' (\u03A3X:', s_Dz,')')
    print('Ly = ', -Ly, ' (\u03A3X:', s_Ly,')')
    print('Lz + Dt = ', -Lz, ' (\u03A3X:', s_Lz,')')
      

     
if __name__ == "__main__" : 
      
    main()
